#include <string.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <allegro.h>
#include <aldumb.h>
#include "datafile.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

//Colours
#define MENU_COL 2
#define MENU_HL_COL 3
#define BLACK 1
#define GREY 2
#define WHITE 3

#define IRND(n) (rand()%(n))
#define FRND(x) (rand()*((x)/(RAND_MAX+1.0f)))

//Data
static DATAFILE *dat;
static COLOR_MAP *firecolmap[16];
static COLOR_MAP *heatcolmap;
static AL_DUH_PLAYER *dp=NULL;
static float dp_refvolume=0;	//the volume before our music slider was incorporated

static int sfxvol=16, musicvol=16;

//Parameters to playmusic()
#define MENUMUSIC 0, 1.5f
#define GAMEMUSIC 32, 1.0f

static void playmusic(int order, float volume) {
	//Stop previous music if any
	al_stop_duh(dp);
	//Start new music at the specified order
	dp_refvolume=volume;
	DUH_SIGRENDERER *sr=dumb_it_start_at_order((DUH *)dat[THE_MUSIC].dat,2,order);
	dp=al_duh_encapsulate_sigrenderer(sr,volume*musicvol/16,4096,44100);
	if (!dp) duh_end_sigrenderer(sr);
}

static void updatemusicvolume() {
	al_duh_set_volume(dp,dp_refvolume*musicvol/16);
}

static void playtestsound() {
	stop_sample((SAMPLE *)dat[SFX_BOOM].dat);
	play_sample((SAMPLE *)dat[SFX_BOOM].dat,128*sfxvol/16,128,1000,0);
}

//Timing
//------

//truetime - writes are only ever done by timerfunc
static volatile int truetime=0;
//logictime - writes are only ever done by main thread
static volatile int logictime=0;

static bool quitgame=false;
static bool fullscreen;
static bool changefullscreen=false;

static BITMAP *scrbuf;

static void timerfunc() {
	if (truetime<logictime+10) truetime++;
}
END_OF_STATIC_FUNCTION(timerfunc);

//Called whenever a game-freezing operation has occurred
static void doneupdate() {
	logictime=truetime;
}

//Current state
//-------------

#define MS_MENU 0
#define MS_FLIGHT 1
#define MS_WORLD 2
#define MS_ENTERNAME 3
#define MS_CHANGEFULLSCREEN 4
#define MS_QUIT 5

static int mainstate=MS_MENU;

//High score table
//----------------

#define TABLENAMEBYTES 21
static char tablenames[10][TABLENAMEBYTES];
static int tablescores[10];
static int yourlastscore=0;
static int yourscorerow=10;

static void loadtable() {
	char nameid[]="name0", scoreid[]="score0";
	for (int i=0; i<10; i++) {
		nameid[4]=scoreid[5]='0'+i;
		const char *str=get_config_string("highscores", nameid, "");
		strncpy(tablenames[i], str, TABLENAMEBYTES-1);
		tablenames[i][TABLENAMEBYTES-1]=0;
		tablescores[i]=get_config_int("highscores", scoreid, 0);
	}
}

static void savetable() {
	char nameid[]="name0", scoreid[]="score0";
	for (int i=0; i<10; i++) {
		if (tablescores[i]>0) {
			nameid[4]=scoreid[5]='0'+i;
			set_config_string("highscores", nameid, tablenames[i]);
			set_config_int("highscores", scoreid, tablescores[i]);
		}
	}
}

//Background starfield
//--------------------

#define NUMSTARS 200
#define STARBORDER (40*256)

static int starx[NUMSTARS], stary[NUMSTARS];
static int stardist[NUMSTARS];
static int starxv, staryv;

//Star velocity transitions
static int srcstarxv, srcstaryv;
static int deststarxv, deststaryv;
static int startranspos=-1;
static int maxstartranspos;

static void initstars() {
	for (int i=0; i<NUMSTARS; i++) {
		starx[i]=IRND(640*256+STARBORDER*2)-STARBORDER;
		stary[i]=IRND(480*256+STARBORDER*2)-STARBORDER;
		stardist[i]=1+IRND(16);
	}
}

static void setstarspeed(int xv, int yv, int time) {
	srcstarxv=starxv;
	srcstaryv=staryv;
	deststarxv=xv;
	deststaryv=yv;
	startranspos=0;
	maxstartranspos=time;
}

static void updatestars() {
	if (startranspos>=0) {
		startranspos++;
		starxv=srcstarxv+(deststarxv-srcstarxv)*startranspos/maxstartranspos;
		staryv=srcstaryv+(deststaryv-srcstaryv)*startranspos/maxstartranspos;
		if (startranspos==maxstartranspos) startranspos=-1;
	}
	for (int i=0; i<NUMSTARS; i++) {
		starx[i]+=starxv/stardist[i];
		stary[i]+=staryv/stardist[i];
		if (starx[i]<-STARBORDER) {
			starx[i]=640*256+STARBORDER-1;
			stary[i]=IRND(480*256+STARBORDER*2)-STARBORDER;
			stardist[i]=1+IRND(16);
		}
		if (starx[i]>=640*256+STARBORDER) {
			starx[i]=-STARBORDER;
			stary[i]=IRND(480*256+STARBORDER*2)-STARBORDER;
			stardist[i]=1+IRND(16);
		}
		if (stary[i]<-STARBORDER) {
			starx[i]=IRND(640*256+STARBORDER*2)-STARBORDER;
			stary[i]=480*256+STARBORDER-1;
			stardist[i]=1+IRND(16);
		}
		if (stary[i]>=480*256+STARBORDER) {
			starx[i]=IRND(640*256+STARBORDER*2)-STARBORDER;
			stary[i]=-STARBORDER;
			stardist[i]=1+IRND(16);
		}
	}
}

static void drawstars() {
	//Draw an aura around the brightest stars
	for (int i=0; i<NUMSTARS; i++) {
		int rad=3*256/stardist[i];
		if (rad>=256) {
			int level=255*(rad-256)/(3*256);
			int col=makecol(level,level,level);
			int xo=starxv*4/stardist[i], yo=staryv*4/stardist[i];
			int x1=(starx[i]-xo)>>8, x2=(starx[i]+xo)>>8;
			int y1=(stary[i]-yo)>>8, y2=(stary[i]+yo)>>8;
			line(scrbuf, x1-1, y1, x2-1, y2, col);
			line(scrbuf, x1+1, y1, x2+1, y2, col);
			line(scrbuf, x1, y1-1, x2, y2-1, col);
			line(scrbuf, x1, y1+1, x2, y2+1, col);
		}
	}
	//Now draw the middles of the stars
	for (int i=0; i<NUMSTARS; i++) {
		int level=3*256/stardist[i];
		if (level>255) level=255;
		int col=makecol(level,level,level);
		int xo=starxv*4/stardist[i], yo=staryv*4/stardist[i];
		int x1=(starx[i]-xo)>>8, x2=(starx[i]+xo)>>8;
		int y1=(stary[i]-yo)>>8, y2=(stary[i]+yo)>>8;
		line(scrbuf, x1, y1, x2, y2, col);
	}
}

//World
//-----

#define FACESIZE 4
#define TEXSIZE 128
#define FIRESPRITESIZE 64

//World tiles
#define SEA            0 //blocks stuff
#define DESERT         1 //doesn't burn, but accommodates mines
#define MINED_DESERT   2 //explodes
#define FOREST         3 //burns
#define MINED_FOREST   4 //explodes
#define BUILDINGS      5 //burns and gives you points
#define POWERPLANT     6 //explodes
#define FUELDEPOT      7 //really explodes
#define NUMNORMALTILES 8
#define BURNT_DESERT   8
#define BURNT_FOREST   9
#define BURNT_BUILT   10
#define TOTALNUMTILES 11

//These are set in 'worldheat'
#define EXPLOSION_BIT 128
#define FIRE_BITS     96
#define FIRE_UNIT     32
#define HEAT_BITS     31

#define NUM_VISUAL_HEAT_LEVELS 4

static unsigned char worldtiles[6][FACESIZE][FACESIZE];
static unsigned char worldheat[6][FACESIZE][FACESIZE];

#define MAXWORLDTRANSPOS 50
static int worldtranspos=0;

//Look angles, and rate of change variables
static int sideangle,vertangle;
static int sideanglev,vertanglev;

//Mouse is hovering over this point in the world, as of last draw
static int worldmouseface=-1, worldmousex=-1, worldmousey=-1;

#define FPT_FIRE     0
#define FPT_EXP      1
#define FPT_EXPSPAWN 2

struct FIREPARTICLE {
	FIREPARTICLE *next;
	//Coordinates are in model space
	float x,y,z;
	float xv,yv,zv;
	float r;
	int type;
	int totaltime,timeleft;
};

static FIREPARTICLE *fireparticles=NULL;
static int fireupdates=0;

//Separate list used for shooting animation in cutscene
static FIREPARTICLE *shootparticles=NULL;

static void dropfireparticles() {
	while (fireparticles!=NULL) {
		FIREPARTICLE *p=fireparticles;
		fireparticles=p->next;
		free(p);
	}
	while (shootparticles!=NULL) {
		FIREPARTICLE *p=shootparticles;
		shootparticles=p->next;
		free(p);
	}
}

static void gettilecentre(int face, int x, int y, int *ppx, int *ppy, int *ppz) {
	int px,py,pz;
	if (face==0) {px=x*2+1-FACESIZE; py=y*2+1-FACESIZE; pz=-FACESIZE;}	//Front
	else if (face==1) {px=FACESIZE; py=y*2+1-FACESIZE; pz=x*2+1-FACESIZE;}	//Right
	else if (face==2) {px=FACESIZE-x*2-1; py=y*2+1-FACESIZE; pz=FACESIZE;}	//Back
	else if (face==3) {px=-FACESIZE; py=y*2+1-FACESIZE; pz=FACESIZE-x*2-1;}	//Left
	else if (face==4) {px=x*2+1-FACESIZE; py=-FACESIZE; pz=FACESIZE-y*2-1;}	//Top
	else {px=x*2+1-FACESIZE; py=FACESIZE; pz=y*2+1-FACESIZE;}	//Bottom
	//Project to unit sphere
	int mag=(int)sqrt(px*px+py*py+pz*pz);
	*ppx=px*256/mag; *ppy=py*256/mag; *ppz=pz*256/mag;
}

static void getexactsurfacepoint(int face, float x, float y, float *ppx, float *ppy, float *ppz) {
	float px,py,pz;
	if (face==0) {px=x-FACESIZE/2; py=y-FACESIZE/2; pz=-FACESIZE/2;}	//Front
	else if (face==1) {px=FACESIZE/2; py=y-FACESIZE/2; pz=x-FACESIZE/2;}	//Right
	else if (face==2) {px=FACESIZE/2-x; py=y-FACESIZE/2; pz=FACESIZE/2;}	//Back
	else if (face==3) {px=-FACESIZE/2; py=y-FACESIZE/2; pz=FACESIZE/2-x;}	//Left
	else if (face==4) {px=x-FACESIZE/2; py=-FACESIZE/2; pz=FACESIZE/2-y;}	//Top
	else {px=x-FACESIZE/2; py=FACESIZE/2; pz=y-FACESIZE/2;}	//Bottom
	//Project to unit sphere
	float mag=256/(float)sqrt(px*px+py*py+pz*pz);
	*ppx=px*mag; *ppy=py*mag; *ppz=pz*mag;
}

static void getadjacent(int dir, int face, int x, int y, int *aface, int *ax, int *ay) {
	if (dir==0) {
		x--;
		if (x<0) {
			if (face==4) {face=3; x=y; y=0;}	//Top to left
			else if (face==5) {face=3; x=FACESIZE-1-y; y=FACESIZE-1;}	//Bottom to left
			else {x=FACESIZE-1; face=(face+3)%4;}	//Leftwards round the equatorial band
		}
	}
	if (dir==1) {
		x++;
		if (x>=FACESIZE) {
			if (face==4) {face=1; x=FACESIZE-1-y; y=0;}	//Top to right
			else if (face==5) {face=1; x=y; y=FACESIZE-1;}	//Bottom to right
			else {x=0; face=(face+1)%4;}	//Rightwards round the equatorial band
		}
	}
	if (dir==2) {
		y--;
		if (y<0) {
			if (face==0) {face=4; x=x; y=FACESIZE-1;}	//Front to top
			else if (face==1) {face=4; y=FACESIZE-1-x; x=FACESIZE-1;}	//Right to top
			else if (face==2) {face=4; x=FACESIZE-1-x; y=0;}	//Back to top
			else if (face==3) {face=4; y=x; x=0;}	//Left to top
			else if (face==4) {face=2; x=FACESIZE-1-x; y=0;}	//Top to back
			else {face=0; x=x; y=FACESIZE-1;}	//Bottom to front
		}
	}
	if (dir==3) {
		y++;
		if (y>=FACESIZE) {
			if (face==0) {face=5; x=x; y=0;}	//Front to bottom
			else if (face==1) {face=5; y=x; x=FACESIZE-1;}	//Right to bottom
			else if (face==2) {face=5; x=FACESIZE-1-x; y=FACESIZE-1;}	//Back to bottom
			else if (face==3) {face=5; y=FACESIZE-1-x; x=0;}	//Left to bottom
			else if (face==4) {face=0; x=x; y=0;}	//Top to front
			else {face=2; x=FACESIZE-1-x; y=FACESIZE-1;}	//Bottom to back
		}
	}
	*aface=face; *ax=x; *ay=y;
}

static void buildworld() {
	worldtranspos=0;
	//Temperature always starts at zero everywhere
	memset(worldheat,0,sizeof(worldheat));
	//Clear worldtiles to 128 to start with
	memset(worldtiles,128,sizeof(worldtiles));
	//Choose some random planes, and increment/decrement tiles on either side
	for (int i=0; i<15; i++) {
		int xc=IRND(257)-128, yc=IRND(257)-128, zc=IRND(257)-128;
		int xn, yn, zn;
		while (true) {
			xn=IRND(257)-128; yn=IRND(257)-128; zn=IRND(257)-128;
			if ((xn|yn|zn)==0) continue; else break;
		}
		for (int face=0; face<6; face++) {
			for (int y=0; y<FACESIZE; y++) {
				for (int x=0; x<FACESIZE; x++) {
					int px,py,pz;
					gettilecentre(face,x,y,&px,&py,&pz);
					px-=xc; py-=yc; pz-=zc;
					if (px*xn+py*yn+pz*zn>=0) worldtiles[face][y][x]++;
					else worldtiles[face][y][x]--;
				}
			}
		}
	}
	//Stuff which has increased overall becomes land - everything else becomes sea
	//First check we'll have land in decent proportions
	int forestlevel=133, landlevel=129;
	while (true) {
		int forestcount=0, landcount=6*FACESIZE*FACESIZE;
		for (int face=0; face<6; face++) {
			for (int y=0; y<FACESIZE; y++) {
				for (int x=0; x<FACESIZE; x++) {
					if (worldtiles[face][y][x]>=forestlevel) forestcount++;
					else if (worldtiles[face][y][x]>=landlevel) {}
					else landcount--;
				}
			}
		}
		bool repeat=false;
		if (forestcount<6*FACESIZE*FACESIZE/6) {forestlevel--; repeat=true;}
		if (landcount<6*FACESIZE*FACESIZE/3) {landlevel--; repeat=true;}
		if (repeat) continue; else break;
	}
	//Now generate the forest, desert and sea
	int landcount=6*FACESIZE*FACESIZE;
	for (int face=0; face<6; face++) {
		for (int y=0; y<FACESIZE; y++) {
			for (int x=0; x<FACESIZE; x++) {
				if (worldtiles[face][y][x]>=forestlevel) worldtiles[face][y][x]=FOREST;
				else if (worldtiles[face][y][x]>=landlevel) worldtiles[face][y][x]=DESERT;
				else {worldtiles[face][y][x]=SEA; landcount--;}
			}
		}
	}
	//Create three cities
	int fillqueue[9];
	for (int i=0; i<3; i++) {
		int cityseed=IRND(landcount);
		int pos=0;
		for (int face=0; face<6; face++) {
			for (int y=0; y<FACESIZE; y++) {
				for (int x=0; x<FACESIZE; x++) {
					if (worldtiles[face][y][x]==FOREST || worldtiles[face][y][x]==DESERT) {
						if (pos==cityseed) {
							int start=0,end=0,max=5+i*2;
							fillqueue[end++]=(face<<16)+(y<<8)+x; worldtiles[face][y][x]=(i==0?FUELDEPOT:POWERPLANT); landcount--;
							while (start<end) {
								int packed=fillqueue[start++];
								for (int dir=0; dir<4; dir++) {
									if (IRND(200)<100) continue;
									getadjacent(dir, packed>>16, packed&255, (packed>>8)&255, &face, &x, &y);
									if (worldtiles[face][y][x]==FOREST || worldtiles[face][y][x]==DESERT) {
										fillqueue[end++]=(face<<16)+(y<<8)+x; worldtiles[face][y][x]=BUILDINGS; landcount--;
										if (end==max) goto done;
									}
								}
							}
							goto done;
						}
						pos++;
					}
				}
			}
		}
		done:;
	}
}

static void unitsphere_and_project(float px, float py, float pz, float *sx, float *sy, float *sz) {
	//Project to unit sphere
	float mag=256.0f/(float)sqrt(px*px+py*py+pz*pz);
	px=px*mag; py=py*mag; pz=pz*mag;
	//Rotate according to player look direction
	float ss=(float)sin(sideangle*(M_PI/256)), cs=(float)cos(sideangle*(M_PI/256));
	float sv=(float)sin(vertangle*(M_PI/256)), cv=(float)cos(vertangle*(M_PI/256));
	float rx=pz*ss+px*cs;
	float ry=(pz*cs-px*ss)*sv+py*cv;
	float rz=(pz*cs-px*ss)*cv-py*sv;
	//Place in front of the camera
	rz+=2000;
	//Project into screen coordinates
	int rad=worldtranspos*1500/MAXWORLDTRANSPOS;
	*sx=320+0.5f+rx*rad/rz;
	*sy=240+0.5f+ry*rad/rz;
	*sz=rz;
}

//Returns true if the sprite is in front of the earth (for sort order purposes)
static bool project_sprite(float px, float py, float pz, float prad, float *sx, float *sy, float *sr) {
	//Rotate according to player look direction
	float ss=(float)sin(sideangle*(M_PI/256)), cs=(float)cos(sideangle*(M_PI/256));
	float sv=(float)sin(vertangle*(M_PI/256)), cv=(float)cos(vertangle*(M_PI/256));
	float rx=pz*ss+px*cs;
	float ry=(pz*cs-px*ss)*sv+py*cv;
	float rz=(pz*cs-px*ss)*cv-py*sv;
	//Place in front of the camera
	rz+=2000;
	//Project into screen coordinates
	int rad=worldtranspos*1500/MAXWORLDTRANSPOS;
	*sx=320+0.5f+rx*rad/rz;
	*sy=240+0.5f+ry*rad/rz;
	*sr=prad*rad/rz;
	//Determine whether it's in front of the earth
	//rx,ry,rz-2000 is the vector from the centre of the earth to this sprite
	//rx,ry,rz is the vector from the camera to this sprite
	//If these two vectors dotted are negative, we're in front
	return rx*rx+ry*ry+(rz-2000)*rz<0;
}

static void createexplosionfire(int numspawns, int numtotal, int px, int py, int pz) {
	for (int i=0; i<numtotal; i++) {
		//Choose a random direction.
		//For points on a sphere, each coordinate is uniformly distributed when considered by itself -
		//so choose z that way, then choose an angle to help get x and y!
		float zv=FRND(1.98f)-0.99f;
		float a=FRND((float)M_PI), r=(float)sqrt(1.0f-zv*zv);
		float xv=r*(float)sin(a), yv=r*(float)cos(a);
		//Scale it by a random velocity - and negate it if necessary to make sure it points out of the globe
		r=(float)sqrt(FRND(20));
		if (xv*px+yv*py+zv*pz<0) r=-r;
		xv*=r; yv*=r; zv*=r;
		//Build the particle
		FIREPARTICLE *p=(FIREPARTICLE *)malloc(sizeof(FIREPARTICLE));
		p->next=fireparticles;
		fireparticles=p;
		p->x=px; p->y=py; p->z=pz;
		p->xv=xv; p->yv=yv; p->zv=zv;
		r=3+FRND(2);
		r=r*r;
		p->r=r;
		if (i<numspawns) {
			p->type=FPT_EXPSPAWN;
			p->totaltime=10+IRND(20);
		}
		else {
			p->type=FPT_EXP;
			p->totaltime=20+IRND(20);
		}
		p->timeleft=p->totaltime;
		//Scale totaltime up in order to darken everything - and darken big ones more
		p->totaltime=(int)(p->totaltime*r/9+0.5f);
	}
}

static void doexplosionfire() {
	int explosioncount=0, firecount=0;
	for (int face=0; face<6; face++) {
		for (int y=0; y<FACESIZE; y++) {
			for (int x=0; x<FACESIZE; x++) {
				if ((worldheat[face][y][x]&EXPLOSION_BIT)!=0) {
					//Get this face's surface coordinates
					int px,py,pz;
					gettilecentre(face,x,y,&px,&py,&pz);
					//Generate a hemisphere of explosion particles starting from this point
					createexplosionfire(4,100,px,py,pz);
					explosioncount++;
				}
				if ((worldheat[face][y][x]&FIRE_BITS)!=0) firecount++;
			}
		}
	}
	//Sounds!
	int vol=(int)sqrt(128*128*explosioncount);
	if (vol>255) vol=255;
	play_sample((SAMPLE *)dat[SFX_BOOM].dat,vol*sfxvol/16,128,950+IRND(101),0);
	vol=(int)sqrt(32*32*firecount);
	if (vol>255) vol=255;
	play_sample((SAMPLE *)dat[SFX_FIRE].dat,vol*sfxvol/16,128,500+IRND(501),0);
}

static void doslowfire() {
	for (int face=0; face<6; face++) {
		for (int y=0; y<FACESIZE; y++) {
			for (int x=0; x<FACESIZE; x++) {
				if ((worldheat[face][y][x]&FIRE_BITS)!=0) {
					//Get a random point on this tile's surface
					//(actually slighty raised in the middle but close enough)
					float px,py,pz;
					getexactsurfacepoint(face,x+FRND(1),y+FRND(1),&px,&py,&pz);
					//Choose a random point on a sphere, for direction randomisation
					//For points on a sphere, each coordinate is uniformly distributed when considered by itself -
					//so choose z that way, then choose an angle to help get x and y!
					float zv=FRND(1.98f)-0.99f;
					float a=FRND((float)M_PI), r=(float)sqrt(1.0f-zv*zv);
					float xv=r*(float)sin(a), yv=r*(float)cos(a);
					xv*=0.5f; yv*=0.5f; zv*=0.5f;
					//Add a randomised velocity in the outward direction
					//(Use px,py,pz which are normalised to 256)
					r=0.5f/256+FRND(0.5f/256);
					xv+=px*r; yv+=py*r; zv+=pz*r;
					//Build the particle
					FIREPARTICLE *p=(FIREPARTICLE *)malloc(sizeof(FIREPARTICLE));
					p->next=fireparticles;
					fireparticles=p;
					p->x=px; p->y=py; p->z=pz;
					p->xv=xv; p->yv=yv; p->zv=zv;
					r=5+FRND(2);
					r=r*r;
					p->r=r;
					p->type=FPT_FIRE;
					p->totaltime=10+IRND(10);
					p->timeleft=p->totaltime;
					//Scale totaltime up in order to darken everything - and darken big ones more
					p->totaltime=(int)(p->totaltime*8*r/25+0.5f);
				}
			}
		}
	}
}

static void createshootexplosion(int numspawns, int numtotal, int px, int py, int pz);

//Since this is CPU-intensive and purely cosmetic, we do it in draw()
//and roughly simulate the number of frames since the last draw()
static void updatefireparticles(int framecount, bool cutscene) {
	//In-game fire particles
	FIREPARTICLE **pp=(cutscene?&shootparticles:&fireparticles);
	FIREPARTICLE *p=*pp;
	float damp=(float)pow(0.9f,framecount);
	while (p!=NULL) {
		p->x+=p->xv*framecount;
		p->y+=p->yv*framecount;
		p->z+=p->zv*framecount;
		if (p->type!=FPT_FIRE) {
			p->xv*=damp;
			p->yv*=damp;
			p->zv*=damp;
		}
		p->timeleft-=framecount;
		if (p->timeleft<=0) {
			//Carefully make sure we never include the new particles in the processing this time round
			//- do this by using the existing p->next - because createexplosionfire() might modify *pp
			FIREPARTICLE *next=p->next;
			*pp=next;
			if (p->type==FPT_EXPSPAWN) {
				if (cutscene) createshootexplosion(0,30,p->x,p->y,p->z);
				else createexplosionfire(0,30,p->x,p->y,p->z);
			}
			free(p);
			p=next;
		}
		else {
			pp=&p->next;
			p=*pp;
		}
	}
}

static void drawfireparticles(bool drawingfront) {
	V3D_f v[4];
	v[0].u=0.01; v[0].v=0.01;
	v[1].u=0.01; v[1].v=FIRESPRITESIZE-0.01;
	v[2].u=FIRESPRITESIZE-0.01; v[2].v=FIRESPRITESIZE-0.01;
	v[3].u=FIRESPRITESIZE-0.01; v[3].v=0.01;
	for (FIREPARTICLE *p=fireparticles; p!=NULL; p=p->next) {
		float sx,sy,sr;
		bool isfront=project_sprite(p->x, p->y, p->z, p->r, &sx, &sy, &sr);
		if (drawingfront!=isfront) continue;
		v[0].x=sx-sr; v[0].y=sy-sr;
		v[1].x=sx-sr; v[1].y=sy+sr;
		v[2].x=sx+sr; v[2].y=sy+sr;
		v[3].x=sx+sr; v[3].y=sy-sr;
		color_map=firecolmap[(p->timeleft*15+p->totaltime-1)/p->totaltime];
		quad3d_f(scrbuf, POLYTYPE_ATEX_MASK_TRANS, (BITMAP *)dat[GFX_FIREPARTICLE].dat, &v[0], &v[1], &v[2], &v[3]);
	}
}

static bool mouseinsideline(int mx, int my, V3D_f *v1, V3D_f *v2) {
	//Basic formula is: mouse_x > v1->x+(v2->x-v1->x)*(mouse_y-v1->y)/(v2->y-v1->y);
	//and this works when v2->y-v1->y is positive
	//but we rearrange it to avoid division, then it magically works for horizontal lines
	//and the other case too - since we multiply through by (v2->y-v1->y)
	return (mx-v1->x)*(v2->y-v1->y) > (v2->x-v1->x)*(my-v1->y);
}

static bool mouseinsidetriangle(int mx, int my, V3D_f *v1, V3D_f *v2, V3D_f *v3) {
	if (!mouseinsideline(mx,my,v1,v2)) return false;
	if (!mouseinsideline(mx,my,v2,v3)) return false;
	if (!mouseinsideline(mx,my,v3,v1)) return false;
	return true;
}

static bool rendertriangle(V3D_f *v1, V3D_f *v2, V3D_f *v3, BITMAP *tex) {
	//Only draw it if it's facing towards the camera (i.e. points go anticlockwise)
	//(Our coordinate system seems to be opposite what Allegro expects,
	//so we have to invert the comparison from what Allegro documents)
	if (polygon_z_normal_f(v1,v2,v3)<=0) {
		triangle3d_f(scrbuf, POLYTYPE_PTEX_LIT, tex, v1, v2, v3);
		//Make a note of whether the mouse is over this triangle
		//Try some different mouse positions in order to make sure we don't leave gaps
		int mp=mouse_pos;
		int mx=mp>>16, my=(short)mp;
		if (mouseinsidetriangle(mx, my, v1, v2, v3)) return true;
		if (mouseinsidetriangle(mx+1, my, v1, v2, v3)) return true;
		if (mouseinsidetriangle(mx, my+1, v1, v2, v3)) return true;
		if (mouseinsidetriangle(mx+1, my+1, v1, v2, v3)) return true;
	}
	return false;
}

//if holdmousestate is true, we don't update worldmouseface etc.
static void drawworld(bool domousehover, bool holdmousestate) {
	V3D_f v[4];
	int hovercoords[8];
	v[0].u=0.01; v[0].v=0.01;
	v[1].u=0.01; v[1].v=TEXSIZE-0.01;
	v[2].u=TEXSIZE-0.01; v[2].v=TEXSIZE-0.01;
	v[3].u=TEXSIZE-0.01; v[3].v=0.01;
	drawfireparticles(false);
	color_map=heatcolmap;
	for (int face=0; face<6; face++) {
		for (int y=0; y<FACESIZE; y++) {
			for (int x=0; x<FACESIZE; x++) {
				int tile=worldtiles[face][y][x];
				int heat=worldheat[face][y][x]&HEAT_BITS;
				if (heat>NUM_VISUAL_HEAT_LEVELS) heat=NUM_VISUAL_HEAT_LEVELS;
				v[0].c=v[1].c=v[2].c=v[3].c=heat;
				BITMAP *tex=(BITMAP *)dat[TILE_000+tile].dat;
				if (face==0) {
					//Front
					v[0].x=x-FACESIZE/2; v[0].y=y-FACESIZE/2; v[0].z=-FACESIZE/2;
					v[1].x=x-FACESIZE/2; v[1].y=y+1-FACESIZE/2; v[1].z=-FACESIZE/2;
					v[2].x=x+1-FACESIZE/2; v[2].y=y+1-FACESIZE/2; v[2].z=-FACESIZE/2;
					v[3].x=x+1-FACESIZE/2; v[3].y=y-FACESIZE/2; v[3].z=-FACESIZE/2;
				}
				else if (face==1) {
					//Right
					v[0].x=FACESIZE/2; v[0].y=y-FACESIZE/2; v[0].z=x-FACESIZE/2;
					v[1].x=FACESIZE/2; v[1].y=y+1-FACESIZE/2; v[1].z=x-FACESIZE/2;
					v[2].x=FACESIZE/2; v[2].y=y+1-FACESIZE/2; v[2].z=x+1-FACESIZE/2;
					v[3].x=FACESIZE/2; v[3].y=y-FACESIZE/2; v[3].z=x+1-FACESIZE/2;
				}
				else if (face==2) {
					//Back
					v[0].x=FACESIZE/2-x; v[0].y=y-FACESIZE/2; v[0].z=FACESIZE/2;
					v[1].x=FACESIZE/2-x; v[1].y=y+1-FACESIZE/2; v[1].z=FACESIZE/2;
					v[2].x=FACESIZE/2-x-1; v[2].y=y+1-FACESIZE/2; v[2].z=FACESIZE/2;
					v[3].x=FACESIZE/2-x-1; v[3].y=y-FACESIZE/2; v[3].z=FACESIZE/2;
				}
				else if (face==3) {
					//Left
					v[0].x=-FACESIZE/2; v[0].y=y-FACESIZE/2; v[0].z=FACESIZE/2-x;
					v[1].x=-FACESIZE/2; v[1].y=y+1-FACESIZE/2; v[1].z=FACESIZE/2-x;
					v[2].x=-FACESIZE/2; v[2].y=y+1-FACESIZE/2; v[2].z=FACESIZE/2-x-1;
					v[3].x=-FACESIZE/2; v[3].y=y-FACESIZE/2; v[3].z=FACESIZE/2-x-1;
				}
				else if (face==4) {
					//Top
					v[0].x=x-FACESIZE/2; v[0].y=-FACESIZE/2; v[0].z=FACESIZE/2-y;
					v[1].x=x-FACESIZE/2; v[1].y=-FACESIZE/2; v[1].z=FACESIZE/2-y-1;
					v[2].x=x+1-FACESIZE/2; v[2].y=-FACESIZE/2; v[2].z=FACESIZE/2-y-1;
					v[3].x=x+1-FACESIZE/2; v[3].y=-FACESIZE/2; v[3].z=FACESIZE/2-y;
				}
				else {
					//Bottom
					v[0].x=x-FACESIZE/2; v[0].y=FACESIZE/2; v[0].z=y-FACESIZE/2;
					v[1].x=x-FACESIZE/2; v[1].y=FACESIZE/2; v[1].z=y+1-FACESIZE/2;
					v[2].x=x+1-FACESIZE/2; v[2].y=FACESIZE/2; v[2].z=y+1-FACESIZE/2;
					v[3].x=x+1-FACESIZE/2; v[3].y=FACESIZE/2; v[3].z=y-FACESIZE/2;
				}
				unitsphere_and_project(v[0].x, v[0].y, v[0].z, &v[0].x, &v[0].y, &v[0].z);
				unitsphere_and_project(v[1].x, v[1].y, v[1].z, &v[1].x, &v[1].y, &v[1].z);
				unitsphere_and_project(v[2].x, v[2].y, v[2].z, &v[2].x, &v[2].y, &v[2].z);
				unitsphere_and_project(v[3].x, v[3].y, v[3].z, &v[3].x, &v[3].y, &v[3].z);
				bool hover=false;
				if (x/(FACESIZE/2)==y/(FACESIZE/2)) {
					//top left or bottom right of face - triangulate by creating edge from point 0 to 2
					hover|=rendertriangle(&v[0], &v[1], &v[2], tex);
					hover|=rendertriangle(&v[2], &v[3], &v[0], tex);
				}
				else {
					//top right or bottom left of face - triangulate by creating edge from point 1 to 3
					hover|=rendertriangle(&v[1], &v[2], &v[3], tex);
					hover|=rendertriangle(&v[3], &v[0], &v[1], tex);
				}
				//Record mouse hover
				if (hover && domousehover && !holdmousestate) {
					worldmouseface=face; worldmousex=x; worldmousey=y;
				}
				if (domousehover && worldmouseface==face && worldmousex==x && worldmousey==y) {
					hovercoords[0]=(int)v[0].x; hovercoords[1]=(int)v[0].y;
					hovercoords[2]=(int)v[1].x; hovercoords[3]=(int)v[1].y;
					hovercoords[4]=(int)v[2].x; hovercoords[5]=(int)v[2].y;
					hovercoords[6]=(int)v[3].x; hovercoords[7]=(int)v[3].y;
				}
			}
		}
	}
	//If we've found a mouse hover, overlay a white rectangle
	if (domousehover && worldmouseface>=0) {
		for (int y=0; y<2; y++) {
			for (int x=0; x<2; x++) {
				line(scrbuf, hovercoords[0]+x, hovercoords[1]+y, hovercoords[2]+x, hovercoords[3]+y, WHITE);
				line(scrbuf, hovercoords[2]+x, hovercoords[3]+y, hovercoords[4]+x, hovercoords[5]+y, WHITE);
				line(scrbuf, hovercoords[4]+x, hovercoords[5]+y, hovercoords[6]+x, hovercoords[7]+y, WHITE);
				line(scrbuf, hovercoords[6]+x, hovercoords[7]+y, hovercoords[0]+x, hovercoords[1]+y, WHITE);
			}
		}
	}
	drawfireparticles(true);
}

static void toworld();

//Space flight
//------------

#define MAXSHIPTRANSOUTPOS 100
static int shiptransoutpos=MAXSHIPTRANSOUTPOS;

#define maxalientransoutpos 300
static int alientransoutpos=maxalientransoutpos;

static const char *preworld1[]={
	"Right. You there. Yes, you, with the ship.",
	"You're mine.",
	"NEWPAGE",
	"Yes, that's right.",
	"For the rest of eternity, you get to do my bidding.",
	"Aren't you lucky!",
	"NEWPAGE",
	"Because I'm a sad, sad, sadistic person ...",
	"... with a stammer ...",
	"... I want the following 7 alien planets destroyed.",
	"NEWPAGE",
	"You have 10 mines and 7 missiles.",
	"That's one missile per planet,",
	"and you can make use of the mines as you see fit.",
	"NEWPAGE",
	"Aha, I see you've nearly arrived. Do enjoy yourself.",
	"END"
};
static const char *preworld2[]={
	"You bloody twit.",
	"NEWPAGE",
	"There were so many things wrong with that attack,",
	"I don't know where to begin.",
	"NEWPAGE",
	"Sea and desert do not burn. Forest does.",
	"Use the forest to spread your explosions.",
	"NEWPAGE",
	"I trust I won't have to remind you.",
	"NEWPAGE",
	"Second world coming right up.",
	"END"
};
static const char *preworld3[]={
	"Fool.",
	"NEWPAGE",
	"At this rate you'll run out of mines.",
	"I mean if you haven't already.",
	"Or you'll end up with some left over that you",
	"could have used to get a higher score.",
	"NEWPAGE",
	"Welcome to the third world.",
	"Let's see what you can do here.",
	"NEWPAGE",
	"Racist.",
	"END"
};
static const char *preworld4[]={
	"Why do I even bother?",
	"You're probably getting sick of me anyway.",
	"Here's World 4. Go on, sod off.",
	"END"
};
static const char *preworld5[]={
	"I'm placing you on performance review.",
	"NEWPAGE",
	"If I don't see a vast improvement over the next three worlds,",
	"I'm afraid I will have no choice but to terminate your employment.",
	"NEWPAGE",
	"All right, we're at World 5.",
	"This had better be something damn special.",
	"Noob.",
	"END"
};
static const char *preworld6[]={
	"Good, now you're getting it.",
	"That was much better.",
	"NEWPAGE",
	"I mean, the skill with which you placed the mines just right",
	"and used the blasts of the power plants and the fuel depot",
	"to maximise the destruction was excellent. I'm impressed!",
	"NEWPAGE",
	"Approaching World 6.",
	"Keep up the good work!",
	"END"
};
static const char *preworld7[]={
	"You know, I've just about had it with you.",
	"NEWPAGE",
	"I try everything.",
	"I try insulting you, I try threatening you,",
	"I even try lying through my teeth and complimenting you,",
	"and you just don't get it.",
	"NEWPAGE",
	"Maybe it's me.",
	"Maybe you find it difficult to work for a",
	"man who has Teenage Catgirls in Heat on DVD.",
	"NEWPAGE",
	"It's not porn. It just has an unfortunate title.",
	"NEWPAGE",
	"And I've stopped saying oishii when I eat bento.",
	"NEWPAGE",
	"I say umai now.",
	"NEWPAGE",
	"Though I do wonder, what's the point?",
	"No one ever talks to me.",
	"NEWPAGE",
	"What? Emo? Yes I bloody well am emo.",
	"Look at me, look at all my problems!",
	"NEWPAGE",
	"Oh I'm so depressed, and no one cares,",
	"someone acknowledge my existence plox!",
	"NEWPAGE",
	"I know an emo pirate too.",
	"NEWPAGE",
	"She has hair over one eye ...",
	"NEWPAGE",
	"... and an eye patch over the other eye.",
	"NEWPAGE",
	"That must be why she likes me.",
	"She can't actually see me.",
	"NEWPAGE",
	"But that doesn't change the fact that I'm in charge",
	"and you are required to do as I say.",
	"NEWPAGE",
	"What? You're at World 7 already?",
	"I haven't finished with you yet!",
	"Where are you going? Get back here!",
	"You arrogant, good-for-nothing, stuck-up piece of sh-",
	"END"
};
static const char *aliensfightback[]={
	"Aaaaaaahahahahahahahahahahahaha.",
	"Everything is falling into place.",
	"SHIPTEXT",
	"Whuh?",
	"NEWPAGE",
	"Hello aliens.",
	"SHIPTEXT",
	"What are all those",
	"aliens doing here?",
	"ALIENTEXT",
	"Omae wa orera no",
	"sekai wo BOMU shita zo.",
	"NEWPAGE",
	"Bomb shit? Ahaha zeeky-boogy-doog boom!",
	"SHIPTEXT",
	"What are they saying?",
	"SHOOT",
	"CHOICE",
	"What the hell are you doing?!",
	"Hey, I just varnished that!",
	"Eh? Seven on one? Come on!",
	"ANSWER", "NEWPAGE", "They seem to be taking their rage", "against my evil deeds out on you.",
	"ANSWER", "NEWPAGE", "You blithering idiot!", "I only bought you that varnish", "to entice you to join my evil scheme.",
	"ANSWER", "NEWPAGE", "You think this is supposed to be fair?", "That's right, you go ahead and despair there.",
		"I'm not a bloody Care Bear.", "NEWPAGE", "And YOU, entheh, will you frigging stop it", "with the frigging rhymes?",
		"Evil villains don't rhyme", "in these frigging times.", "NEWPAGE", "... Gah!",
	"ENDCHOICE",
	"SHOOT",
	"CHOICE",
	"Argh! Wait! Stop shooting!",
	"Look, it's a cosmic tear!",
	"I love your ships. Are they from Sunny Place Station?",
	"ANSWER", "ALIENTEXT", "Nanda yo, kono", "henna uchuujin-go!", "Zenzen wakannee zo!",
		"NEWPAGE", "Yeah, I agree! How dare you even ask!",
	"ANSWER", "ALIENTEXT", "E? Doko? Tte iu ka", "orera wo damasu ki ka!", "NEWPAGE", "Eh? How? He just said those are pure fiction!",
	"ANSWER", "ALIENTEXT", "Ara, kininatta no!", "Sou yo ... tte yokumo", "baka koto kiku ki ka!",
		"NEWPAGE", "Yeah, he does indeed! This bastard ...", "is not fit to fly flocking fag fleets!",
	"ENDCHOICE",
	"SHIPTEXT", "You speak their language?", "NEWPAGE", "Err no!", "Well just a smattering.", "SHIPTEXT", ">:I",
	"NEWPAGE", "All right, I studied it for four years", "when I was a young lad around your age.", "These guys are old friends.",
	"SHOOT",
	"CHOICE",
	"Wait, amarillion made me do it!",
	"It wasn't me! We all have ships like this!",
	"Stop! Look, I'll blow my own planet up, how's that?",
	"ANSWER", "ALIENTEXT", "Arumadiro?", "Doko da soitsu?", "NEWPAGE", "I'm on Earth.", "Err I mean he's on Earth. Too.",
		"ALIENTEXT", "Sokka. Koitsu ni", "fukushuu wo toru ni wa", "shi ga amai.", "ALIENTEXT", "Aa, koitsu ni ataeru",
		"yotei no KURUSHIMI no", "koto wo omou to", "moeru ze, kyuiin!", "SHIPTEXT", "What did they say?",
		"NEWPAGE", "I don't think you wanna know.", "SHIPTEXT", "Oh come on, tell me!",
		"NEWPAGE", "Death is too good for this tosser,", "and I quote: \"Oh, thinking about the",
		"TERROR we plan to inflict on him", "makes me burn sexually, waaawwww.\"", "NEWPAGE", "I've seen these guys.",
		"They have tentacles.", "SHIPTEXT", "My god.", "We have to get there", "first and kill everyone.", "NEWPAGE", "Yeah.",
	"ANSWER", "ALIENTEXT", "Sou iu nara,", "shou ga nai.", "Kono ... AASU ... wo", "hakai to sekai seiha da.",
		"NEWPAGE", "They said that if what you say is true,", "then there's no choice.", "They're going to have to destroy",
		"and world-dominate your arse.", "SHIPTEXT", "Can't they find some other moon?",
		"ALIENTEXT", "Chau nen!", "AASU! Chikyuu!", "NEWPAGE", "Oh. They meant Earth.", "They lack the vocal anatomy necessary",
		"to differentiate 'arse' and 'Earth'.", "SHIPTEXT", "I like these guys.", "I'll help them.", "ALIENTEXT",
		"Maa ikka.", "NEWPAGE", "They seem taken with the idea.", "Well whatever, I always wanted", "to blow a planet up.",
	"ANSWER", "NEWPAGE", "GASP!", "ALIENTEXT", "GASUPU!", "SHIPTEXT", "Gleeg snag zip!", "SHIPTEXT", "...",
		"NEWPAGE", "What is this?", "SHIPTEXT", "Fuck, it didn't work.", "I'll go and blow it up now, OK?",
		"ALIENTEXT", "Aa. Sou suru ga ii.", "NEWPAGE", "Old pal, it seems we haven't got a choice.",
		"NEWPAGE", "We can always go and live on a beach on Saturn.", "NEWPAGE", "I once dreamt there were beautiful girls there.",
		"NEWPAGE", "I was too young to want to rape them back then.", "NEWPAGE", "...", "NEWPAGE", "Yeah, let's do it!",
	"ENDCHOICE",
	"END"
};

/*
Upon seeing 'CHOICE', display the choices and wait for a number to be pressed - then display the text
- proceed to the relevant ANSWER - only parse 'ANSWER' when not in a nested CHOICE/ENDCHOICE
- upon seeing another ANSWER, skip to an ENDCHOICE (when not in a nested CHOICE/ENDCHOICE)
- upon seeing an ENDCHOICE, revert to normal mode
*/

static const char **gamemessages[]={
	preworld1,
	preworld2,
	preworld3,
	preworld4,
	preworld5,
	preworld6,
	preworld7,
	aliensfightback
};

static const char **currentpage;
static int numtextlines;
static int numchars;
static int textdelay;
static char tmpstr[256];
static int textx,texty;

#define CS_NONE 0
#define CS_INPUT 1
#define CS_SHOWING 2
static int choicestate;

static int shootframe;

static int mines, missiles;

static void startflight() {
	dropfireparticles();
	mainstate=MS_FLIGHT;
	float angle=(7-missiles)*M_PI/7;
	if ((missiles&1)!=0) angle=-angle;
	setstarspeed((int)(-20*256*cos(angle)),(int)(20*256*sin(angle)),100);
	currentpage=gamemessages[7-missiles];
	numtextlines=0;
	numchars=0;
	choicestate=CS_NONE;
	shootframe=0;
	textdelay=2;
	textx=(missiles==0?220:320); texty=400;
}

static void createshootexplosion(int numspawns, int numtotal, int px, int py, int pz) {
	for (int i=0; i<numtotal; i++) {
		//Choose a random direction.
		//For points on a sphere, each coordinate is uniformly distributed when considered by itself -
		//so choose z that way, then choose an angle to help get x and y!
		float zv=FRND(1.98f)-0.99f;
		float a=FRND((float)M_PI), r=(float)sqrt(1.0f-zv*zv);
		float xv=r*(float)sin(a), yv=r*(float)cos(a);
		//Scale it by a random velocity
		r=(float)sqrt(FRND(20));
		xv*=r; yv*=r; zv*=r;
		//Build the particle
		FIREPARTICLE *p=(FIREPARTICLE *)malloc(sizeof(FIREPARTICLE));
		p->next=shootparticles;
		shootparticles=p;
		p->x=px; p->y=py; p->z=pz;
		p->xv=xv; p->yv=yv; p->zv=zv;
		r=3+FRND(2);
		r=r*r;
		p->r=r;
		if (i<numspawns) {
			p->type=FPT_EXPSPAWN;
			p->totaltime=10+IRND(20);
		}
		else {
			p->type=FPT_EXP;
			p->totaltime=20+IRND(20);
		}
		p->timeleft=p->totaltime;
		//Scale totaltime up in order to darken everything - and darken big ones more
		p->totaltime=(int)(p->totaltime*r/9+0.5f);
	}
}

static void updateflight() {
	//Update shoot particles if any - deferred to drawing function
	fireupdates++;
	//Shooting animation active?
	if (shootframe>0) {
		int myx=160+40, myy=240;
		//Create fire particles along lines from aliens to player
		if (shootframe<15) {
			for (int i=0; i<7; i++) {
				int alienx=460+80*i*(6-i)/9;
				int yspacing=195;
				int alieny=240+(i-3)*yspacing/3;
				int firex=alienx+(myx-alienx)*(shootframe+2)/17;
				int firey=alieny+(myy-alieny)*(shootframe+2)/17;
				//Choose a random point on a sphere, for direction randomisation
				//For points on a sphere, each coordinate is uniformly distributed when considered by itself -
				//so choose z that way, then choose an angle to help get x and y!
				float zv=FRND(1.98f)-0.99f;
				float a=FRND((float)M_PI), r=(float)sqrt(1.0f-zv*zv);
				float xv=r*(float)sin(a), yv=r*(float)cos(a);
				xv*=0.5f; yv*=0.5f; zv*=0.5f;
				//Build the particle
				FIREPARTICLE *p=(FIREPARTICLE *)malloc(sizeof(FIREPARTICLE));
				p->next=shootparticles;
				shootparticles=p;
				p->x=firex; p->y=firey; p->z=1000;
				p->xv=xv; p->yv=yv; p->zv=zv;
				r=2+FRND(1);
				r=r*r;
				p->r=r;
				p->type=FPT_FIRE;
				p->totaltime=10+IRND(10);
				p->timeleft=p->totaltime;
			}
		}
		//Create the explosion at the player
		if (shootframe==15) {
			createshootexplosion(4,100,myx,myy,1000);
			play_sample((SAMPLE *)dat[SFX_BOOM].dat,255*sfxvol/16,128,1000,0);
		}
		//Next frame next time
		shootframe++;
		if (shootframe==50) shootframe=0;
		//Do nothing else while shooting animation is active
		return;
	}
	//Do nothing while waiting for answer to a choice
	if (choicestate==CS_INPUT) return;
	if (textdelay>0) {
		textdelay--;
		if (textdelay==0) {
			textdelay=2;
			if (currentpage[numtextlines][numchars]!=0) {
				numchars++;
				if (currentpage[numtextlines][numchars]==0) {
					textdelay=20;
					if (choicestate>=CS_SHOWING) textdelay=50;
					else {
						const char *next=currentpage[numtextlines+1];
						if (strcmp(next,"NEWPAGE")==0) textdelay=50;
						if (strcmp(next,"SHIPTEXT")==0) textdelay=50;
						if (strcmp(next,"ALIENTEXT")==0) textdelay=50;
						if (strcmp(next,"CHOICE")==0) textdelay=50;
						if (strcmp(next,"ANSWER")==0) textdelay=50;
						if (strcmp(next,"ENDCHOICE")==0) textdelay=50;
						if (strcmp(next,"END")==0) textdelay=50;
						if (strcmp(next,"SHOOT")==0) textdelay=50;
					}
				}
			}
			else {
				//If a choice was showing, then skip to the relevant ANSWER block now
				if (choicestate>=CS_SHOWING) {
					//Skip to the relevant 'ANSWER' block (skipping any nested choices)
					int nest=0;
					while (true) {
						currentpage++;
						if (nest==0 && strcmp(*currentpage,"ANSWER")==0) {
							if (choicestate==CS_SHOWING) break; else choicestate--;
						}
						if (strcmp(*currentpage,"CHOICE")==0) nest++;
						if (strcmp(*currentpage,"ENDCHOICE")==0) nest--;
					}
					choicestate=CS_NONE;
				}
				//Normally we don't loop ('break' at the end)
				//but we do loop in a few cases - e.g. when we've processed ENDCHOICE,
				//we always want to process the directive after it immediately
				while (true) {
					const char *next=currentpage[numtextlines+1];
					if (strcmp(next,"NEWPAGE")==0) {
						textx=(missiles==0?220:320); texty=400;
						currentpage+=numtextlines+2;
						numtextlines=0; numchars=0;
					}
					else if (strcmp(next,"SHIPTEXT")==0 || strcmp(next,"CHOICE")==0) {
						textx=160; texty=280;
						currentpage+=numtextlines+2;
						numtextlines=0; numchars=0;
						if (strcmp(next,"CHOICE")==0) choicestate=CS_INPUT;
					}
					else if (strcmp(next,"ALIENTEXT")==0) {
						textx=388; texty=224;
						currentpage+=numtextlines+2;
						numtextlines=0; numchars=0;
					}
					else if (strcmp(next,"ANSWER")==0) {
						currentpage+=numtextlines+1;
						//Skip to the next 'ENDCHOICE' block (skipping any nested choices)
						int nest=0;
						while (true) {
							currentpage++;
							if (nest==0 && strcmp(*currentpage,"ENDCHOICE")==0) break;
							if (strcmp(*currentpage,"CHOICE")==0) nest++;
							if (strcmp(*currentpage,"ENDCHOICE")==0) nest--;
						}
						numtextlines=0;
						continue;
					}
					else if (strcmp(next,"ENDCHOICE")==0) {
						currentpage+=numtextlines+1;
						numtextlines=0;
						continue;
					}
					else if (strcmp(next,"END")==0) {
						toworld();
					}
					else if (strcmp(next,"SHOOT")==0) {
						currentpage+=numtextlines+1;
						numtextlines=0;
						shootframe=1;
						play_sample((SAMPLE *)dat[SFX_ROAR].dat,128*sfxvol/16,128,1000,0);
						continue;
					}
					else {
						numtextlines++; numchars=0;
					}
					break;
				}
			}
		}
	}
}

static void drawshootparticles(bool drawingfront) {
	V3D_f v[4];
	v[0].u=0.01; v[0].v=0.01;
	v[1].u=0.01; v[1].v=FIRESPRITESIZE-0.01;
	v[2].u=FIRESPRITESIZE-0.01; v[2].v=FIRESPRITESIZE-0.01;
	v[3].u=FIRESPRITESIZE-0.01; v[3].v=0.01;
	for (FIREPARTICLE *p=shootparticles; p!=NULL; p=p->next) {
		bool isfront=(p->z<=1000);
		if (drawingfront!=isfront) continue;
		//Project into screen coordinates
		float sx=320+0.5f+(p->x-320)*1000/p->z;
		float sy=240+0.5f+(p->y-240)*1000/p->z;
		float sr=p->r*1000/p->z;
		v[0].x=sx-sr; v[0].y=sy-sr;
		v[1].x=sx-sr; v[1].y=sy+sr;
		v[2].x=sx+sr; v[2].y=sy+sr;
		v[3].x=sx+sr; v[3].y=sy-sr;
		color_map=firecolmap[(p->timeleft*15+p->totaltime-1)/p->totaltime];
		quad3d_f(scrbuf, POLYTYPE_ATEX_MASK_TRANS, (BITMAP *)dat[GFX_FIREPARTICLE].dat, &v[0], &v[1], &v[2], &v[3]);
	}
}

static void drawflight() {
	drawshootparticles(false);
	//Draw the ships
	if (shiptransoutpos<MAXSHIPTRANSOUTPOS || alientransoutpos<maxalientransoutpos) {
		//Draw your ship
		int offset=480*2*shiptransoutpos*shiptransoutpos/(MAXSHIPTRANSOUTPOS*MAXSHIPTRANSOUTPOS);
		//Ship always faces right - keep xv positive
		int xv=deststarxv, yv=deststaryv;
		if (xv==0 && yv==0) {xv=srcstarxv; yv=srcstaryv;}
		if (missiles!=0) {xv=-xv; yv=-yv;}	//fly backwards when out of missiles (ending sequence)
		float mag=(float)sqrt(xv*xv+yv*yv);
		//Ship always comes on from the left
		int x=320-(int)(offset*xv/mag);
		int y=240-(int)(offset*yv/mag);
		//Slide ship to the left if the alien ships are coming
		x=x-160+160*alientransoutpos*alientransoutpos/(maxalientransoutpos*maxalientransoutpos);
		BITMAP *spr=(BITMAP *)dat[SHIP_000+(logictime&4)/4].dat;
		float angle=(float)atan2(yv, xv);	//HA!
		//rotate_sprite is just gonna convert it back to vectors anyway though :P
		if (xv<0) rotate_sprite_v_flip(scrbuf, spr, x-spr->w/2, y-spr->h/2, (int)(angle*(65536*128/M_PI)));
		else rotate_sprite(scrbuf, spr, x-spr->w/2, y-spr->h/2, (int)(angle*(65536*128/M_PI)));
		//Draw the alien ships
		int myx=x, myy=y;
		if (alientransoutpos<maxalientransoutpos) {
			for (int i=0; i<7; i++) {
				x=460+80*i*(6-i)/9;
				int yspacing=195+60*alientransoutpos*alientransoutpos/(maxalientransoutpos*maxalientransoutpos);
				y=240+(i-3)*yspacing/3;
				x+=(240+480*i*(6-i)/9)*alientransoutpos*alientransoutpos/(maxalientransoutpos*maxalientransoutpos);
				//Make the ship face towards the player
				BITMAP *spr=(BITMAP *)dat[ALIENSHIP_000+i*2+(logictime&4)/4].dat;
				float angle=(float)atan2(y-myy, x-myx);	//muahaha, comment above applies :b
				rotate_sprite(scrbuf, spr, x-spr->w/2, y-spr->h/2, (int)(angle*(65536*128/M_PI)));
			}
		}
	}
	drawshootparticles(true);
	//Draw text captions
	if (mainstate==MS_FLIGHT) {
		if (shootframe==0) {
			if (choicestate==CS_INPUT) {
				textout_ex(scrbuf,font,"Press a number to choose:",30,texty,WHITE,-1);
				for (int i=0; ; i++) {
					const char *choice=currentpage[i];
					if (strcmp(choice,"ANSWER")==0) break;
					textprintf_ex(scrbuf,font,30,texty+11+i*11,WHITE,-1,"%d. %s",i+1,choice);
				}
			}
			else {
				for (int i=0; i<numtextlines; i++) {
					int x=textx-text_length(font,currentpage[i])/2;
					int y=texty+i*11;
					if (x<10) x=10;
					textout_ex(scrbuf,font,currentpage[i],x,y,WHITE,-1);
				}
				int x=textx-text_length(font,currentpage[numtextlines])/2;
				int y=texty+numtextlines*11;
				if (x<10) x=10;
				memcpy(tmpstr,currentpage[numtextlines],numchars);
				strcpy(tmpstr+numchars,(logictime&4)!=0?"_":"");
				textout_ex(scrbuf,font,tmpstr,x,y,WHITE,-1);
			}
		}
		//Draw skip instruction
		if ((logictime&32)!=0) textout_right_ex(scrbuf,font,"Space/Enter to skip",640-4,480-12,WHITE,-1);
	}
}

static void stopgame();

static void handleflightkey(int k) {
	if (k>>8==KEY_ESC) {stopgame(); return;}
	if (k>>8==KEY_SPACE || k>>8==KEY_ENTER) {
		play_sample((SAMPLE *)dat[SFX_ACTIVATE].dat,255*sfxvol/16,128,1000,0);
		toworld(); return;
	}
	if (shootframe==0 && choicestate==CS_INPUT) {
		for (int i=0; ; i++) {
			const char *choice=currentpage[i];
			if (strcmp(choice,"ANSWER")==0) break;
			if ((k&255)=='1'+i) {
				play_sample((SAMPLE *)dat[SFX_ACTIVATE].dat,255*sfxvol/16,128,1000,0);
				currentpage+=i; choicestate=CS_SHOWING+i; return;
			}
		}
	}
}

//Game
//----

static bool specifyattack=false;
static int specifyattacktranspos=0;
#define maxspecifyattacktranspos 15

static const char *attackerror;
static int attackerrortime=0;

static int nextpropagatein=0;
#define PROPAGATEPERIOD 15
static bool propagatefinished;

static int minesmissilestranspos=0;
#define maxminesmissilestranspos 20

static int subtitletranspos=0;
#define maxsubtitletranspos MAXWORLDTRANSPOS

static int leftscoretranspos=0;
#define maxleftscoretranspos 10
static int rightscoretranspos=0;
#define maxrightscoretranspos 20

static int totalscore;

static int numbuildings, numforest, numpowerplants, numfueldepots;
static bool allbuildings, allravaged;

#define BUILDINGSCORE 100
#define FORESTSCORE 10
#define POWERPLANTSCORE 400
#define FUELDEPOTSCORE 1000
#define ALLBUILDINGS 500
#define ALLRAVAGED 500

static int levelscore() {
	int score=numbuildings*BUILDINGSCORE+numforest*FORESTSCORE+numpowerplants*POWERPLANTSCORE+numfueldepots*FUELDEPOTSCORE;
	if (allbuildings) score+=ALLBUILDINGS;
	if (allravaged) score+=ALLRAVAGED;
	return score;
}

static void toworld() {
	dropfireparticles();
	specifyattack=false;
	attackerrortime=0;
	nextpropagatein=0;
	propagatefinished=false;
	mainstate=MS_WORLD;
	buildworld();
	setstarspeed(0,0,100);
	totalscore+=levelscore();
	numbuildings=numforest=numpowerplants=numfueldepots=0;
	allbuildings=allravaged=false;
}

static void showattackerror(const char *text) {
	attackerror=text;
	attackerrortime=50;
	specifyattack=true;
	play_sample((SAMPLE *)dat[SFX_FAIL].dat,255*sfxvol/16,128,1000,0);
}

static void startgame() {
	mines=10; missiles=7;
	totalscore=0;
	numbuildings=numforest=numpowerplants=numfueldepots=0;
	allbuildings=allravaged=false;
	startflight();
	playmusic(GAMEMUSIC);
}

static void stopgame() {
	dropfireparticles();
	setstarspeed(0,0,100);
	mainstate=MS_MENU;
	playmusic(MENUMUSIC);
	//Add me to the high score table?
	yourlastscore=totalscore+levelscore();
	yourscorerow=10;
	while (yourlastscore>0 && yourscorerow>0 && yourlastscore>=tablescores[yourscorerow-1]) {
		if (yourscorerow<10) {
			tablescores[yourscorerow]=tablescores[yourscorerow-1];
			memcpy(&tablenames[yourscorerow], &tablenames[yourscorerow-1], TABLENAMEBYTES);
		}
		yourscorerow--;
	}
	if (yourscorerow<10) {
		tablenames[yourscorerow][0]=0;
		tablescores[yourscorerow]=yourlastscore;
		mainstate=MS_ENTERNAME;
	}
}

static void addheat(int face, int x, int y, int dir, int amount) {
	if (dir>=0) getadjacent(dir,face,x,y,&face,&x,&y);
	int packed=worldheat[face][y][x];
	int heat=packed&HEAT_BITS;
	heat+=amount;
	if (heat>HEAT_BITS) heat=HEAT_BITS;
	worldheat[face][y][x]=(packed&~HEAT_BITS)+heat;
}

static void setburnmark(int face, int x, int y) {
	int tile=worldtiles[face][y][x];
	if (tile==DESERT) worldtiles[face][y][x]=BURNT_DESERT;
	if (tile==FOREST) {numforest++; worldtiles[face][y][x]=BURNT_FOREST;}
	if (tile==BUILDINGS) {numbuildings++; worldtiles[face][y][x]=BURNT_BUILT;}
}

static void sufferheat(int face, int x, int y);

static void explode(int face, int x, int y, int dir) {
	if (dir>=0) getadjacent(dir,face,x,y,&face,&x,&y);
	worldheat[face][y][x]|=EXPLOSION_BIT;
	addheat(face,x,y,-1,11);
	addheat(face,x,y,0,2);
	addheat(face,x,y,1,2);
	addheat(face,x,y,2,2);
	addheat(face,x,y,3,2);
	//Respond to heat immediately for the tile under the explosion
	sufferheat(face,x,y);
	//If we haven't already, create an immediate explosion mark here
	setburnmark(face,x,y);
}

static void setfire(int face, int x, int y) {
	int heat=worldheat[face][y][x];
	if ((heat&FIRE_BITS)==0) {
		worldheat[face][y][x]=heat+FIRE_UNIT*3;
		addheat(face,x,y,-1,2);
	}
}

static void sufferheat(int face, int x, int y) {
	int tile=worldtiles[face][y][x];
	if (tile==MINED_DESERT) {worldtiles[face][y][x]=BURNT_DESERT; explode(face,x,y,-1);}
	if (tile==FOREST) setfire(face,x,y);
	if (tile==MINED_FOREST) {numforest++; worldtiles[face][y][x]=BURNT_FOREST; explode(face,x,y,-1); setfire(face,x,y);}
	if (tile==BUILDINGS) setfire(face,x,y);
	if (tile==POWERPLANT) {numpowerplants++; worldtiles[face][y][x]=BURNT_BUILT; explode(face,x,y,-1);}
	if (tile==FUELDEPOT) {
		numfueldepots++;
		worldtiles[face][y][x]=BURNT_BUILT;
		//Explosions in the four adjacent tiles will create enough heat in the centre tile
		//So all we need to do is manually place an explosion animation in the centre
		worldheat[face][y][x]|=EXPLOSION_BIT;
		explode(face,x,y,0);
		explode(face,x,y,1);
		explode(face,x,y,2);
		explode(face,x,y,3);
	}
}

static void propagate() {
	/*
	Phenomena to order:
	1. run through and find all tiles which are already hot, and all those ready to explode
	2. suffer heat and process head spreads at their locations
	3. if burning, increase, else reduce heat at their locations by 1 - and update fire
	- explosion: immediately create fire underneath, just generate heat adjacent (for fire later)
	*/
	//Take a copy of the heat map as it was before any processing
	unsigned char oldheatmap[6][FACESIZE][FACESIZE];
	memcpy(oldheatmap,worldheat,sizeof(worldheat));
	//Remove old explosions
	for (int face=0; face<6; face++) {
		for (int y=0; y<FACESIZE; y++) {
			for (int x=0; x<FACESIZE; x++) {
				worldheat[face][y][x]&=~EXPLOSION_BIT;
			}
		}
	}
	//The following variables will be unset as we notice stuff happening
	propagatefinished=true;
	allbuildings=true;
	allravaged=true;
	//Process the tiles
	for (int face=0; face<6; face++) {
		for (int y=0; y<FACESIZE; y++) {
			for (int x=0; x<FACESIZE; x++) {
				//Suffer heat for everything that's already hot
				int oldheat=oldheatmap[face][y][x];
				if ((oldheat&HEAT_BITS)>0) sufferheat(face,x,y);
				//Spread if hot enough
				int exp=oldheat&HEAT_BITS;
				if (exp>=6) {worldheat[face][y][x]++; exp++;}
				//worldheat may have just overflowed but we're going to undo that now!
				while (exp>=6) {
					worldheat[face][y][x]-=4;
					exp-=4;
					addheat(face,x,y,0,1);
					addheat(face,x,y,1,1);
					addheat(face,x,y,2,1);
					addheat(face,x,y,3,1);
				}
				//Process fire
				if ((oldheat&FIRE_BITS)>0) {
					worldheat[face][y][x]-=FIRE_UNIT;
					if ((worldheat[face][y][x]&FIRE_BITS)==0) setburnmark(face,x,y);
					else addheat(face,x,y,-1,2);
				}
				//Reduce heat over time
				if ((oldheat&HEAT_BITS)>0) worldheat[face][y][x]--;
				//Have we finished?
				if (oldheat!=0) propagatefinished=false;
				int tile=worldtiles[face][y][x];
				if (tile==FOREST || tile==MINED_FOREST) allravaged=false;
				if (tile==BUILDINGS || tile==POWERPLANT || tile==FUELDEPOT) allravaged=allbuildings=false;
			}
		}
	}
}

static void handleworldkey(int k) {
	if (k>>8==KEY_ESC) {
		if (specifyattack) {specifyattack=false; attackerrortime=0;}
		else stopgame();
		return;
	}
	if (nextpropagatein==0 && attackerrortime==0 && k>>8==KEY_M && worldmouseface>=0) {
		attackerrortime=0; attackerror=NULL;
		if (mines<=0) showattackerror("You have no mines left");
		else {
			int tile=worldtiles[worldmouseface][worldmousey][worldmousex];
			if (tile==DESERT) {
				mines--; worldtiles[worldmouseface][worldmousey][worldmousex]=MINED_DESERT;
				play_sample((SAMPLE *)dat[SFX_MINE].dat,255*sfxvol/16,128,1000,0);
			}
			else if (tile==FOREST) {
				mines--; worldtiles[worldmouseface][worldmousey][worldmousex]=MINED_FOREST;
				play_sample((SAMPLE *)dat[SFX_MINE].dat,255*sfxvol/16,128,1000,0);
			}
			else if (tile==MINED_DESERT || tile==MINED_FOREST) showattackerror("A mine has already been placed here");
			else showattackerror("Mines can only be placed in forest or desert");
		}
		if (attackerrortime==0) specifyattack=false;
		return;
	}
	if (nextpropagatein==0 && attackerrortime==0 && k>>8==KEY_SPACE && worldmouseface>=0) {
		attackerrortime=0; attackerror=NULL;
		int tile=worldtiles[worldmouseface][worldmousey][worldmousex];
		if (tile==SEA) showattackerror("Missiles must be launched towards land");
		else {
			play_sample((SAMPLE *)dat[SFX_HIGHLIGHT].dat,255*sfxvol/16,128,1000,0);
			explode(worldmouseface,worldmousex,worldmousey,-1);
			doexplosionfire();
			nextpropagatein=PROPAGATEPERIOD;
			missiles--;
		}
		if (attackerrortime==0) specifyattack=false;
		return;
	}
	if (k>>8==KEY_ENTER || k>>8==KEY_SPACE) {
		if (attackerrortime!=0) {attackerrortime=0; specifyattack=false; return;}
		if (propagatefinished) {
			play_sample((SAMPLE *)dat[SFX_ACTIVATE].dat,255*sfxvol/16,128,1000,0);
			if (missiles>=0) startflight(); else stopgame();
			return;
		}
	}
}

static void updateworld() {
	//Check for click on world
	if (nextpropagatein==0 && !specifyattack && worldmouseface>=0 && (mouse_b&3)!=0) {
		play_sample((SAMPLE *)dat[SFX_HIGHLIGHT].dat,255*sfxvol/16,128,1000,0);
		specifyattack=true;
		attackerrortime=0; attackerror=NULL;
	}
	//Update viewing angle
	if (!specifyattack) {
		int sideanglea=0; if (key[KEY_LEFT] || key[KEY_A]) sideanglea-=6; if (key[KEY_RIGHT] || key[KEY_D]) sideanglea+=6;
		int vertanglea=0; if (key[KEY_UP] || key[KEY_W]) vertanglea-=6; if (key[KEY_DOWN] || key[KEY_S]) vertanglea+=6;
		if (sideanglev<sideanglea) sideanglev++;
		if (sideanglev>sideanglea) sideanglev--;
		if (vertanglev<vertanglea) vertanglev++;
		if (vertanglev>vertanglea) vertanglev--;
		sideangle=(sideangle+sideanglev)&511;
		vertangle+=vertanglev;
		if (vertangle<-64) {vertangle=-64; vertanglev=0;}
		if (vertangle>64) {vertangle=64; vertanglev=0;}
	}
	//Update attack error
	if (attackerrortime>0) {
		attackerrortime--;
		if (attackerrortime==0) specifyattack=false;
	}
	//Process explosions
	if (nextpropagatein>0) {
		nextpropagatein--;
		if (nextpropagatein==0) {
			propagate();
			doexplosionfire();
			nextpropagatein=PROPAGATEPERIOD;
		}
	}
	//Generate slow fire every time round
	doslowfire();
	//Update fire particles (deferred to draw function)
	fireupdates++;
}

//logic and drawing functions
//---------------------------

//Menu state
#define MAXTITLETRANSPOS 10
static int titletranspos=0;
#define MAXTABLETRANSPOS 20
static int tabletranspos=0;
#define MAXMENUTRANSPOS 10
static int menutranspos=0;
#define MENU_PLAY 0
#define MENU_SFXVOL 1
#define MENU_MUSICVOL 2
#define MENU_FULLSCREEN 3
#define MENU_QUIT 4
#define MENU_LENGTH 5
static int menupos=0;

static void init() {
	initstars();
}

static void menu_activateoption() {
	if (menupos==MENU_SFXVOL) playtestsound();
	if (menupos==MENU_PLAY) startgame();
	if (menupos==MENU_FULLSCREEN) mainstate=MS_CHANGEFULLSCREEN;
	if (menupos==MENU_QUIT) mainstate=MS_QUIT;
}

static void logic() {
	//Handle buffered keys
	while (keypressed()) {
		int k=readkey();
		//Menu keys
		if (mainstate==MS_MENU) {
			if (k>>8==KEY_ESC) mainstate=MS_QUIT;
			if (k>>8==KEY_UP) {
				menupos=(menupos+MENU_LENGTH-1)%MENU_LENGTH;
				play_sample((SAMPLE *)dat[SFX_HIGHLIGHT].dat,255*sfxvol/16,128,1000,0);
			}
			if (k>>8==KEY_DOWN) {
				menupos=(menupos+1)%MENU_LENGTH;
				play_sample((SAMPLE *)dat[SFX_HIGHLIGHT].dat,255*sfxvol/16,128,1000,0);
			}
			if (k>>8==KEY_LEFT) {
				if (menupos==MENU_SFXVOL && sfxvol>0) {sfxvol--; playtestsound();}
				if (menupos==MENU_MUSICVOL && musicvol>0) {musicvol--; updatemusicvolume();}
			}
			if (k>>8==KEY_RIGHT) {
				if (menupos==MENU_SFXVOL && sfxvol<16) {sfxvol++; playtestsound();}
				if (menupos==MENU_MUSICVOL && musicvol<16) {musicvol++; updatemusicvolume();}
			}
			if (k>>8==KEY_ENTER || k>>8==KEY_SPACE) {
				menu_activateoption();
				play_sample((SAMPLE *)dat[SFX_ACTIVATE].dat,255*sfxvol/16,128,1000,0);
			}
		}
		//Game keys
		else if (mainstate==MS_FLIGHT) handleflightkey(k);
		else if (mainstate==MS_WORLD) handleworldkey(k);
		//High score name entry keys
		else if (mainstate==MS_ENTERNAME) {
			char *name=tablenames[yourscorerow];
			int len=strlen(name);
			if ((k&255)>=32 && (k&255)!=127) {
				if (len<TABLENAMEBYTES-1) {
					name[len++]=(char)(k&255);
					name[len]=0;
				}
			}
			else if ((k&255)==8 || (k&255)==127) {
				if (len>0) {len--; name[len]=0;}
			}
			else if (k>>8==KEY_ENTER || k>>8==KEY_ESC) {
				savetable();
				mainstate=MS_MENU;
			}
		}
	}
	//Keep stars moving
	updatestars();
	//Update game logic
	if (mainstate==MS_FLIGHT) updateflight();
	if (mainstate==MS_WORLD) updateworld();
	//Handle transitions
	if (mainstate==MS_MENU && tabletranspos==MAXTABLETRANSPOS) {if (menutranspos<MAXMENUTRANSPOS) menutranspos++;}
	else {if (menutranspos>0) menutranspos--;}
	if ((mainstate==MS_MENU || mainstate==MS_ENTERNAME) && shiptransoutpos>=MAXSHIPTRANSOUTPOS/2 && worldtranspos==0) {
		if (titletranspos<MAXTITLETRANSPOS) titletranspos++;
	}
	else {if (titletranspos>0) titletranspos--;}
	if ((mainstate==MS_MENU || mainstate==MS_ENTERNAME) && titletranspos==MAXTITLETRANSPOS) {
		if (tabletranspos<MAXTABLETRANSPOS) tabletranspos++;
	}
	else {if (tabletranspos>0) tabletranspos--;}
	if (mainstate==MS_WORLD && shiptransoutpos==MAXSHIPTRANSOUTPOS) {if (worldtranspos<MAXWORLDTRANSPOS) worldtranspos++;}
	else {if (worldtranspos>0) worldtranspos--;}
	if (mainstate==MS_FLIGHT) {if (shiptransoutpos>0) shiptransoutpos--;}
	else {if (shiptransoutpos<MAXSHIPTRANSOUTPOS) shiptransoutpos++;}
	if (mainstate==MS_FLIGHT && missiles==0) {if (alientransoutpos>0) alientransoutpos--;}
	else {alientransoutpos+=3; if (alientransoutpos>maxalientransoutpos) alientransoutpos=maxalientransoutpos;}
	if (specifyattack) {if (specifyattacktranspos<maxspecifyattacktranspos) specifyattacktranspos++;}
	else {if (specifyattacktranspos>0) specifyattacktranspos--;}
	if (mainstate==MS_WORLD || mainstate==MS_FLIGHT) {if (minesmissilestranspos<maxminesmissilestranspos) minesmissilestranspos++;}
	else {if (minesmissilestranspos>0) minesmissilestranspos--;}
	if (mainstate==MS_WORLD && shiptransoutpos==MAXSHIPTRANSOUTPOS && nextpropagatein==0) {
		if (subtitletranspos<maxsubtitletranspos) subtitletranspos++;
	}
	else {if (subtitletranspos>0) subtitletranspos--;}
	if (mainstate==MS_WORLD && nextpropagatein>0) {if (leftscoretranspos<maxleftscoretranspos) leftscoretranspos++;}
	else {if (leftscoretranspos>0) leftscoretranspos--;}
	if (mainstate==MS_WORLD && nextpropagatein>0 && propagatefinished) {
		if (rightscoretranspos==0) play_sample((SAMPLE *)dat[SFX_WHOOSH].dat,32*sfxvol/16,128,1000,0);
		if (rightscoretranspos<maxrightscoretranspos) rightscoretranspos++;
	}
	else {if (rightscoretranspos>0) rightscoretranspos--;}
	if (mainstate==MS_QUIT && menutranspos==0) quitgame=true;
	if (mainstate==MS_CHANGEFULLSCREEN && menutranspos==0) {changefullscreen=true; mainstate=MS_MENU;}
}

static void draw() {
	clear_bitmap(scrbuf);
	drawstars();
	//Draw the various parts of the flight view as necessary
	drawflight();
	//Draw the world
	bool domousehover=(mainstate==MS_WORLD && nextpropagatein==0);
	bool holdmousestate=specifyattack;
	if (!holdmousestate) {worldmouseface=-1; worldmousex=-1; worldmousey=-1;}
	if (fireupdates>0) {updatefireparticles(fireupdates,false); updatefireparticles(fireupdates,true); fireupdates=0;}
	if (worldtranspos>0) drawworld(domousehover,holdmousestate);
	if (leftscoretranspos>0) {
		int boxoffset=27*8-27*8*leftscoretranspos*leftscoretranspos/(maxleftscoretranspos*maxleftscoretranspos);
		rect(scrbuf,-1-boxoffset,409,26*8-boxoffset,470,WHITE);
		rectfill(scrbuf,0-boxoffset,410,26*8-1-boxoffset,469,BLACK);
		textprintf_ex(scrbuf,font,4+16*8-13*8-boxoffset,414,WHITE,-1,"Buildings: %d \xD7 %dpts",numbuildings,BUILDINGSCORE);
		textprintf_ex(scrbuf,font,4+16*8-10*8-boxoffset,428,WHITE,-1,"Forest: %d \xD7 %dpts",numforest,FORESTSCORE);
		textprintf_ex(scrbuf,font,4+16*8-16*8-boxoffset,442,WHITE,-1,"Power plants: %d \xD7 %dpts",numpowerplants,POWERPLANTSCORE);
		textprintf_ex(scrbuf,font,4+16*8-15*8-boxoffset,456,WHITE,-1,"Fuel depots: %d \xD7 %dpts",numfueldepots,FUELDEPOTSCORE);
	}
	if (rightscoretranspos>0) {
		int boxoffset=25*8-25*8*rightscoretranspos*rightscoretranspos/(maxrightscoretranspos*maxrightscoretranspos);
		rect(scrbuf,639-24*8+boxoffset,409,640+boxoffset,470,WHITE);
		rectfill(scrbuf,640-24*8+boxoffset,410,639+boxoffset,469,BLACK);
		textprintf_ex(scrbuf,font,640-4-20*8+boxoffset,414,WHITE,-1,"All buildings: %d",allbuildings?ALLBUILDINGS:0);
		textprintf_ex(scrbuf,font,640-4-23*8+boxoffset,428,WHITE,-1,"All land ravaged: %d",allravaged?ALLRAVAGED:0);
		textprintf_ex(scrbuf,font,640-4-18*8+boxoffset,442,WHITE,-1,"Level score: %d",levelscore());
		textprintf_ex(scrbuf,font,640-4-18*8+boxoffset,456,WHITE,-1,"Total score: %d",totalscore+levelscore());
		if (rightscoretranspos==maxrightscoretranspos && (logictime&32)!=0) {
			textout_right_ex(scrbuf,font,"Space/Enter to continue",640-4,396,WHITE,-1);
		}
	}
	if (specifyattacktranspos>0) {
		int boxoffset=640-640*specifyattacktranspos*specifyattacktranspos/(maxspecifyattacktranspos*maxspecifyattacktranspos);
		rect(scrbuf,70+boxoffset,200,569+boxoffset,279,WHITE);
		rectfill(scrbuf,71+boxoffset,201,568+boxoffset,278,BLACK);
		if (attackerror!=NULL) {
			textout_centre_ex(scrbuf,font,attackerror,320+boxoffset,240-4,WHITE,-1);
		}
		else {
			textout_centre_ex(scrbuf,font,"Press M to place a mine here.",320+boxoffset,210,WHITE,-1);
			textout_centre_ex(scrbuf,font,"Press SPACE to launch a missile here.",320+boxoffset,224,WHITE,-1);
			textout_centre_ex(scrbuf,font,"Hint: you don't have to click first. For the next one,",320+boxoffset,246,WHITE,-1);
			textout_centre_ex(scrbuf,font,"position the mouse without clicking, and press M or SPACE.",320+boxoffset,260,WHITE,-1);
		}
	}
	//Draw various text overlays
	if (subtitletranspos>0) {
		int offset=(480-440)-(480-440)*subtitletranspos*subtitletranspos/(maxsubtitletranspos*maxsubtitletranspos);
		textout_centre_ex(scrbuf,font,"Use the arrow keys or WASD to rotate the world.",320,440+offset,WHITE,-1);
		textout_centre_ex(scrbuf,font,"Use the mouse to execute your attack.",320,452+offset,WHITE,-1);
	}
	if (minesmissilestranspos>0) {
		int offset=28-minesmissilestranspos*28/maxminesmissilestranspos;
		textprintf_ex(scrbuf,font,8,8-offset,WHITE,-1,"Score: %d",totalscore+levelscore());
		int world=(nextpropagatein>0?7-missiles:8-missiles);
		if (mainstate==MS_WORLD && world>=8) {
			textout_ex(scrbuf,font,"World: Earth",8,20-offset,WHITE,-1);
			textout_right_ex(scrbuf,font,"Missiles: Found one!",640-8,8-offset,WHITE,-1);
		}
		else {
			if (mainstate==MS_WORLD) {
				textprintf_ex(scrbuf,font,8,20-offset,WHITE,-1,"World: %d",world);
			}
			textprintf_right_ex(scrbuf,font,640-8,8-offset,WHITE,-1,"Missiles: %d",missiles);
		}
		textprintf_right_ex(scrbuf,font,640-8,20-offset,WHITE,-1,"Mines: %d",mines);
	}
	//Draw the title
	if (titletranspos>0) {
		int titleoffset=120-120*titletranspos*titletranspos/(MAXTITLETRANSPOS*MAXTITLETRANSPOS);
		BITMAP *spr=(BITMAP *)dat[GFX_TITLE].dat;
		draw_sprite(scrbuf,spr,320-spr->w/2,40-spr->h/2-titleoffset);
		textout_centre_ex(scrbuf,font,"Written in 72 hours for TINS 2008",320,40+spr->h/2+12-titleoffset,GREY,-1);
		textout_centre_ex(scrbuf,font,"Dedicated to Kait ^_^",320,40+spr->h/2+24-titleoffset,GREY,-1);
	}
	//Draw the high score table
	if (tabletranspos>0) {
		int scoreoffset=440-440*tabletranspos*tabletranspos/(MAXTABLETRANSPOS*MAXTABLETRANSPOS);
		textout_ex(scrbuf,font,"Name",228-scoreoffset,120,WHITE,-1);
		textout_right_ex(scrbuf,font,"Score",440-scoreoffset,120,WHITE,-1);
		for (int i=0; i<10; i++) {
			textprintf_right_ex(scrbuf,font,224-scoreoffset,142+i*14,WHITE,-1,"%d. ",i+1);
			if (tablescores[i]>0) {
				if (mainstate==MS_ENTERNAME && yourscorerow==i && (logictime&8)!=0) {
					textprintf_ex(scrbuf,font,228+scoreoffset,142+i*14,WHITE,-1,"%s_",tablenames[i]);
				}
				else textout_ex(scrbuf,font,tablenames[i],228+scoreoffset,142+i*14,yourscorerow==i?WHITE:GREY,-1);
				textprintf_right_ex(scrbuf,font,440+scoreoffset,142+i*14,yourscorerow==i?WHITE:GREY,-1,"%d",tablescores[i]);
			}
		}
		if (yourscorerow==10 && yourlastscore>0) {
			textout_ex(scrbuf,font,"Your score:",228+scoreoffset,290,WHITE,-1);
			textprintf_right_ex(scrbuf,font,440+scoreoffset,290,WHITE,-1,"%d",yourlastscore);
		}
		if (mainstate==MS_ENTERNAME) {
			textout_centre_ex(scrbuf,font,"New high score - please enter your name above!",320+scoreoffset,290,WHITE,-1);
		}
	}
	//Draw the menu
	if (menutranspos>0) {
		int menuoffset=140-140*menutranspos*menutranspos/(MAXMENUTRANSPOS*MAXMENUTRANSPOS);
		for (int i=0; i<MENU_LENGTH; i++) {
			const char *str=0;
			if (i==MENU_PLAY) str="Play game";
			if (i==MENU_SFXVOL) {
				str=tmpstr;
				strcpy(tmpstr,"Sound: -----------------");
				tmpstr[7+sfxvol]='|';
			}
			if (i==MENU_MUSICVOL) {
				str=tmpstr;
				strcpy(tmpstr,"Music: -----------------");
				tmpstr[7+musicvol]='|';
			}
			if (i==MENU_FULLSCREEN) str=(fullscreen?"Go windowed":"Go full screen");
			if (i==MENU_QUIT) str="Quit";
			textout_centre_ex(scrbuf,font,str,320,340+menuoffset+i*15,menupos==i?MENU_HL_COL:MENU_COL,-1);
		}
		textout_centre_ex(scrbuf,font,"\xA9 2008 Ben 'Bruce \"entheh\" Perry' Davis",320,450+menuoffset,GREY,-1);
	}
	//Draw the mouse cursor
	if (domousehover) {
		int mp=mouse_pos;
		int mx=mp>>16, my=(short)mp;
		hline(scrbuf,mx-10,my,mx+10,WHITE);
		vline(scrbuf,mx,my-10,my+10,WHITE);
	}
	//Done - blit to screen
	blit(scrbuf,screen,0,0,0,0,640,480);
}

//Bootstrap + fullscreen/windowed management
//------------------------------------------

int main() {
	srand(time(NULL));
	set_uformat(U_ASCII);
	set_config_file("aonfire.ini");
	allegro_init();
	atexit(&dumb_exit);
	install_timer();
	install_keyboard();
	install_mouse();
	reserve_voices(64,-1);
	install_sound(DIGI_AUTODETECT,MIDI_NONE,NULL);
	dumb_register_dat_it_quick(DUMB_DAT_IT);
	dat=load_datafile("datafile.dat");
	if (dat==NULL) {
		allegro_message("Unable to load datafile!\n\n"
				"If you downloaded this game as a 'zip' file,\n"
				"you need to right-click the 'zip' file and\n"
				"choose to extract it somewhere -\n"
				" then open the extracted copy instead.\n\n"
				"I hate Windows.\n");
		exit(1);
	}
	RGB *pal=(RGB *)dat[THE_PALETTE].dat;
	//Create a copy of the palette without the explosion colours
	PALETTE palcopy;
	memcpy(palcopy, pal, sizeof(PALETTE));
	for (int c=240; c<256; c++) palcopy[c].r=palcopy[c].g=palcopy[c].b=0;
	//Build a colour table for the heated surface effect - avoid the explosion colours
	rgb_map=(RGB_MAP *)malloc(sizeof(RGB_MAP));
	create_rgb_table(rgb_map, palcopy, NULL);
	heatcolmap=(COLOR_MAP *)malloc(256*(NUM_VISUAL_HEAT_LEVELS+1));
	for (int col=0; col<256; col++) heatcolmap->data[0][col]=col;
	for (int heat=1; heat<NUM_VISUAL_HEAT_LEVELS+1; heat++) {
		for (int col=0; col<256; col++) {
			int srcr=pal[col].r, srcg=pal[col].g, srcb=pal[col].b;
			int destr=63, destg=(srcr+srcg)/2, destb=(srcr+srcb)/2;
			int r=srcr+(destr-srcr)*heat/NUM_VISUAL_HEAT_LEVELS;
			int g=srcg+(destg-srcg)*heat/NUM_VISUAL_HEAT_LEVELS;
			int b=srcb+(destb-srcb)*heat/NUM_VISUAL_HEAT_LEVELS;
			heatcolmap->data[heat][col]=rgb_map->data[r>>1][g>>1][b>>1];
		}
	}
	//Build colour tables for the fire particle blending
	//'drawlevel' is the level of brightness at which the particle is being drawn
	//(This ruse also means I don't have to z-sort the particles, woo)
	for (int drawlevel=0; drawlevel<16; drawlevel++) {
		firecolmap[drawlevel]=(COLOR_MAP *)malloc(256*17);
		//'srcfirepixel' is the colour in the fire sprite - 0 to 16 inclusive
		for (int srcfirepixel=0; srcfirepixel<17; srcfirepixel++) {
			int destfirepixel=srcfirepixel+drawlevel-15;	//maximum 16 inclusive, may be negative
			if (destfirepixel<0) destfirepixel=0;
			//Colours 0-239 get overwritten with raw fire colours
			for (int c=0; c<240; c++) {
				if (destfirepixel>0) firecolmap[drawlevel]->data[srcfirepixel][c]=239+destfirepixel;
				else firecolmap[drawlevel]->data[srcfirepixel][c]=c;
			}
			//Colours 240-255 get added on to to make brighter fire colours
			//- sort of - it's not quite an add - we take the distances from white,
			//and multiply them as if they are in the range 0 to 1,
			//thereby producing a smaller distance from white (and guaranteeing commutativity)
			for (int c=240; c<256; c++) {
				int whitedist1=255-c, whitedist2=16-destfirepixel;
				int newwhitedist=whitedist1*whitedist2/16;
				firecolmap[drawlevel]->data[srcfirepixel][c]=255-newwhitedist;
			}
		}
	}
	fullscreen=(get_config_int("main","fullscreen",0)!=0);
	sfxvol=get_config_int("main","sfxvol",16);
	musicvol=get_config_int("main","musicvol",16);
	loadtable();
	init();
	playmusic(MENUMUSIC);
	bool tryagain=true;
	install_int_ex(timerfunc,BPS_TO_TIMER(50));
	//The purpose of this loop is to be 'continued' whenever we want to change graphics mode
	while (!quitgame) {
		//Set the desired graphics mode
		int gfxmode=(fullscreen?GFX_AUTODETECT_FULLSCREEN:GFX_AUTODETECT_WINDOWED);
		int rv=set_gfx_mode(gfxmode,640,480,0,0);
		if (rv!=0) {
			if (tryagain) {
				fullscreen=!fullscreen;
				tryagain=false;
				continue;
			}
			else break;
		}
		set_palette(pal);
		scrbuf=create_bitmap(640,480);
		doneupdate();
		//Run the menus/game
		while (true) {
			int runto=truetime;
			if (runto>logictime) {
				while (logictime<runto) {
					logic(); logictime++;
					if (quitgame || changefullscreen) goto breakmainloop;
				}
				draw();
			}
			al_poll_duh(dp);
			rest(1);
		}
		breakmainloop:
		destroy_bitmap(scrbuf);
		if (changefullscreen) {
			fullscreen=!fullscreen;
			changefullscreen=false;
		}
	}
	al_stop_duh(dp);
	set_gfx_mode(GFX_TEXT,0,0,0,0);
	set_config_int("main","fullscreen",fullscreen);
	set_config_int("main","sfxvol",sfxvol);
	set_config_int("main","musicvol",musicvol);
	return 0;
}
END_OF_MAIN();
